import is from '@sindresorhus/is';
import { parseRange } from './helpers.js';
import * as errors from './exceptions.js';
export class ExceptionHandler {
    debug = process.env.NODE_ENV !== 'production';
    renderStatusPages = process.env.NODE_ENV === 'production';
    statusPages = {};
    #expandedStatusPages;
    renderers = {
        html: async (error, ctx) => {
            if (this.isDebuggingEnabled(ctx)) {
                const { default: Youch } = await import('youch');
                const html = await new Youch(error, ctx.request.request).toHTML();
                ctx.response.status(error.status).send(html);
                return;
            }
            ctx.response.status(error.status).send(`<p> ${error.message} </p>`);
        },
        json: async (error, ctx) => {
            if (this.isDebuggingEnabled(ctx)) {
                const { default: Youch } = await import('youch');
                const json = await new Youch(error, ctx.request.request).toJSON();
                ctx.response.status(error.status).send(json.error);
                return;
            }
            ctx.response.status(error.status).send({ message: error.message });
        },
        json_api: async (error, ctx) => {
            if (this.isDebuggingEnabled(ctx)) {
                const { default: Youch } = await import('youch');
                const json = await new Youch(error, ctx.request.request).toJSON();
                ctx.response.status(error.status).send(json.error);
                return;
            }
            ctx.response.status(error.status).send({
                errors: [
                    {
                        title: error.message,
                        code: error.code,
                        status: error.status,
                    },
                ],
            });
        },
    };
    reportErrors = true;
    ignoreExceptions = [
        errors.E_HTTP_EXCEPTION,
        errors.E_ROUTE_NOT_FOUND,
        errors.E_CANNOT_LOOKUP_ROUTE,
        errors.E_HTTP_REQUEST_ABORTED,
    ];
    ignoreStatuses = [400, 422, 401];
    ignoreCodes = [];
    #expandStatusPages() {
        if (!this.#expandedStatusPages) {
            this.#expandedStatusPages = Object.keys(this.statusPages).reduce((result, range) => {
                const renderer = this.statusPages[range];
                result = Object.assign(result, parseRange(range, renderer));
                return result;
            }, {});
        }
        return this.#expandedStatusPages;
    }
    #toHttpError(error) {
        const httpError = is.object(error) ? error : new Error(String(error));
        httpError.message = httpError.message || 'Internal server error';
        httpError.status = httpError.status || 500;
        return httpError;
    }
    context(ctx) {
        const requestId = ctx.request.id();
        return requestId
            ? {
                'x-request-id': requestId,
            }
            : {};
    }
    getResponseFormat(ctx) {
        switch (ctx.request.accepts(['html', 'application/vnd.api+json', 'json'])) {
            case 'application/vnd.api+json':
                return 'json_api';
            case 'json':
                return 'json';
            case 'html':
            default:
                return 'html';
        }
    }
    getErrorLogLevel(error) {
        if (error.status >= 500) {
            return 'error';
        }
        if (error.status >= 400) {
            return 'warn';
        }
        return 'info';
    }
    isDebuggingEnabled(_) {
        return this.debug;
    }
    shouldReport(error) {
        if (this.reportErrors === false) {
            return false;
        }
        if (this.ignoreStatuses.includes(error.status)) {
            return false;
        }
        if (error.code && this.ignoreCodes.includes(error.code)) {
            return false;
        }
        if (this.ignoreExceptions.find((exception) => error instanceof exception)) {
            return false;
        }
        return true;
    }
    async report(error, ctx) {
        const httpError = this.#toHttpError(error);
        if (!this.shouldReport(httpError)) {
            return;
        }
        if (typeof httpError.report === 'function') {
            httpError.report(httpError, ctx);
            return;
        }
        const level = this.getErrorLogLevel(httpError);
        ctx.logger.log(level, {
            ...(level === 'error' || level === 'fatal' ? { err: httpError } : {}),
            ...this.context(ctx),
        }, httpError.message);
    }
    async handle(error, ctx) {
        const httpError = this.#toHttpError(error);
        if (typeof httpError.handle === 'function') {
            return httpError.handle(httpError, ctx);
        }
        const statusPages = this.#expandStatusPages();
        if (this.renderStatusPages && statusPages[httpError.status]) {
            return statusPages[httpError.status](httpError, ctx);
        }
        const responseFormat = this.getResponseFormat(ctx);
        return this.renderers[responseFormat](httpError, ctx);
    }
}
