import onFinished from 'on-finished';
import Middleware from '@poppinss/middleware';
import { moduleCaller, moduleImporter } from '@adonisjs/fold';
import { Qs } from '../qs.js';
import debug from '../debug.js';
import { Request } from '../request.js';
import { Response } from '../response.js';
import { Router } from '../router/main.js';
import { HttpContext } from '../http_context/main.js';
import { finalHandler } from './factories/final_handler.js';
import { writeResponse } from './factories/write_response.js';
import { asyncLocalStorage } from '../http_context/local_storage.js';
import { middlewareHandler } from './factories/middleware_handler.js';
export class Server {
    #defaultErrorHandler = {
        report() { },
        handle(error, ctx) {
            ctx.response.status(error.status || 500).send(error.message || 'Internal server error');
        },
    };
    #logger;
    #errorHandler;
    #resolvedErrorHandler = this.#defaultErrorHandler;
    #emitter;
    #app;
    #encryption;
    #config;
    #qsParser;
    #serverMiddlewareStack;
    #router;
    #nodeHttpServer;
    #middleware = [];
    get usingAsyncLocalStorage() {
        return asyncLocalStorage.isEnabled;
    }
    constructor(app, encryption, emitter, logger, config) {
        this.#app = app;
        this.#emitter = emitter;
        this.#config = config;
        this.#logger = logger;
        this.#encryption = encryption;
        this.#qsParser = new Qs(this.#config.qs);
        this.#router = new Router(this.#app, this.#encryption, this.#qsParser);
        this.#createAsyncLocalStore();
        debug('server config: %O', this.#config);
    }
    #createAsyncLocalStore() {
        if (this.#config.useAsyncLocalStorage) {
            debug('creating ALS store for HTTP context');
            asyncLocalStorage.create();
        }
        else {
            asyncLocalStorage.destroy();
        }
    }
    #createServerMiddlewareStack() {
        this.#serverMiddlewareStack = new Middleware();
        this.#middleware.forEach((middleware) => this.#serverMiddlewareStack.add(middleware));
        this.#serverMiddlewareStack.freeze();
        this.#middleware = [];
    }
    #handleRequest(ctx, resolver) {
        return this.#serverMiddlewareStack.runner()
            .errorHandler((error) => {
            this.#resolvedErrorHandler.report(error, ctx);
            return this.#resolvedErrorHandler.handle(error, ctx);
        })
            .finalHandler(finalHandler(this.#router, resolver, ctx))
            .run(middlewareHandler(resolver, ctx))
            .catch((error) => {
            ctx.logger.fatal({ err: error }, 'Exception raised by error handler');
            return this.#defaultErrorHandler.handle(error, ctx);
        })
            .finally(writeResponse(ctx));
    }
    pipeline(middleware) {
        const middlewareStack = new Middleware();
        middleware.forEach((one) => {
            middlewareStack.add(moduleCaller(one, 'handle').toHandleMethod());
        });
        middlewareStack.freeze();
        const stackRunner = middlewareStack.runner();
        return {
            finalHandler(handler) {
                stackRunner.finalHandler(handler);
                return this;
            },
            errorHandler(handler) {
                stackRunner.errorHandler(handler);
                return this;
            },
            run(ctx) {
                return stackRunner.run((handler, next) => {
                    return handler.handle(ctx.containerResolver, ctx, next);
                });
            },
        };
    }
    use(middleware) {
        middleware.forEach((one) => this.#middleware.push(moduleImporter(one, 'handle').toHandleMethod()));
        return this;
    }
    errorHandler(handler) {
        this.#errorHandler = handler;
        return this;
    }
    async boot() {
        debug('booting HTTP server');
        this.#createServerMiddlewareStack();
        this.#router.commit();
        if (this.#errorHandler) {
            if (debug.enabled) {
                debug('using custom error handler "%s"', this.#errorHandler);
            }
            const moduleExports = await this.#errorHandler();
            this.#resolvedErrorHandler = await this.#app.container.make(moduleExports.default);
        }
    }
    setNodeServer(server) {
        this.#nodeHttpServer = server;
    }
    getNodeServer() {
        return this.#nodeHttpServer;
    }
    getRouter() {
        return this.#router;
    }
    createRequest(req, res) {
        return new Request(req, res, this.#encryption, this.#config, this.#qsParser);
    }
    createResponse(req, res) {
        return new Response(req, res, this.#encryption, this.#config, this.#router, this.#qsParser);
    }
    createHttpContext(request, response, resolver) {
        return new HttpContext(request, response, this.#logger.child({ request_id: request.id() }), resolver);
    }
    handle(req, res) {
        const hasRequestListener = this.#emitter.hasListeners('http:request_finished');
        const startTime = hasRequestListener ? process.hrtime() : null;
        const resolver = this.#app.container.createResolver();
        const ctx = this.createHttpContext(this.createRequest(req, res), this.createResponse(req, res), resolver);
        if (startTime) {
            onFinished(res, () => {
                this.#emitter.emit('http:request_finished', {
                    ctx: ctx,
                    duration: process.hrtime(startTime),
                });
            });
        }
        if (this.usingAsyncLocalStorage) {
            return asyncLocalStorage.storage.run(ctx, () => this.#handleRequest(ctx, resolver));
        }
        return this.#handleRequest(ctx, resolver);
    }
}
